import pytest  # type: ignore
from fps_auth.config import AuthConfig, get_auth_config


@pytest.fixture
def auth_mode():
    return "token"


@pytest.fixture
def auth_config(auth_mode):
    yield AuthConfig.parse_obj({"mode": auth_mode})


@pytest.fixture
def config_override(app, auth_config):
    async def override_get_config():
        return auth_config

    app.dependency_overrides[get_auth_config] = override_get_config



import uuid
from typing import Optional

from fastapi_users import schemas
from pydantic import BaseModel


class JupyterUser(BaseModel):
    anonymous: bool = True
    username: str = ""
    name: Optional[str] = None
    color: Optional[str] = None
    avatar: Optional[str] = None
    workspace: str = "{}"
    settings: str = "{}"


class UserRead(schemas.BaseUser[uuid.UUID], JupyterUser):
    pass


class UserCreate(schemas.BaseUserCreate):
    anonymous: bool = True
    username: Optional[str] = None
    name: Optional[str] = None
    color: Optional[str] = None


class UserUpdate(schemas.BaseUserUpdate, JupyterUser):
    pass



import secrets
from pathlib import Path
from typing import AsyncGenerator, List

from fastapi import Depends
from fastapi_users.db import SQLAlchemyBaseOAuthAccountTableUUID  # type: ignore
from fastapi_users.db import (  # type: ignore
    SQLAlchemyBaseUserTableUUID,
    SQLAlchemyUserDatabase,
)
from fps.config import get_config  # type: ignore
from sqlalchemy import Boolean, Column, String, Text  # type: ignore
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine  # type: ignore
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base  # type: ignore
from sqlalchemy.orm import relationship, sessionmaker  # type: ignore

from .config import AuthConfig

auth_config = get_config(AuthConfig)

jupyter_dir = Path.home() / ".local" / "share" / "jupyter"
jupyter_dir.mkdir(parents=True, exist_ok=True)
secret_path = jupyter_dir / "jupyverse_secret"
userdb_path = jupyter_dir / "jupyverse_users.db"

if auth_config.clear_users:
    if userdb_path.is_file():
        userdb_path.unlink()
    if secret_path.is_file():
        secret_path.unlink()

if not secret_path.is_file():
    with open(secret_path, "w") as f:
        f.write(secrets.token_hex(32))

with open(secret_path) as f:
    secret = f.read()


DATABASE_URL = f"sqlite+aiosqlite:///{userdb_path}"
Base: DeclarativeMeta = declarative_base()


class OAuthAccount(SQLAlchemyBaseOAuthAccountTableUUID, Base):
    pass


class User(SQLAlchemyBaseUserTableUUID, Base):
    anonymous = Column(Boolean, default=True, nullable=False)
    email = Column(String(length=32), nullable=False, unique=True)
    username = Column(String(length=32), nullable=True, unique=True)
    name = Column(String(length=32), nullable=True)
    color = Column(String(length=32), nullable=True)
    avatar = Column(String(length=32), nullable=True)
    workspace = Column(Text(), default="{}", nullable=False)
    settings = Column(Text(), default="{}", nullable=False)
    oauth_accounts: List[OAuthAccount] = relationship("OAuthAccount", lazy="joined")


engine = create_async_engine(DATABASE_URL)
Session = sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)


async def create_db_and_tables():
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)


async def get_async_session() -> AsyncGenerator[AsyncSession, None]:
    async with Session() as session:
        yield session


async def get_user_db(session: AsyncSession = Depends(get_async_session)):
    yield SQLAlchemyUserDatabase(session, User, OAuthAccount)


class UserDb:
    async def __aenter__(self):
        self.session = Session()
        session = await self.session.__aenter__()
        return SQLAlchemyUserDatabase(session, User, OAuthAccount)

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        return await self.session.__aexit__(exc_type, exc_value, exc_tb)



from typing import Optional
from uuid import uuid4

from fps.config import PluginModel, get_config  # type: ignore
from fps.hooks import register_config, register_plugin_name  # type: ignore
from pydantic import SecretStr


class AuthConfig(PluginModel):
    client_id: str = ""
    client_secret: SecretStr = SecretStr("")
    redirect_uri: str = ""
    # mode: Literal["noauth", "token", "user"] = "token"
    mode: str = "token"
    token: str = str(uuid4())
    collaborative: bool = False
    global_email: str = "guest@jupyter.com"
    cookie_secure: bool = False  # FIXME: should default to True, and set to False for tests
    clear_users: bool = False
    login_url: Optional[str] = None


def get_auth_config():
    return get_config(AuthConfig)


c = register_config(AuthConfig)
n = register_plugin_name("authenticator")



import uuid
from typing import Generic, Optional

import httpx
from fastapi import Depends, HTTPException, Response, status
from fastapi_users import (  # type: ignore
    BaseUserManager,
    FastAPIUsers,
    UUIDIDMixin,
    models,
)
from fastapi_users.authentication import (
    AuthenticationBackend,
    CookieTransport,
    JWTStrategy,
)
from fastapi_users.authentication.strategy.base import Strategy
from fastapi_users.authentication.transport.base import Transport
from fastapi_users.db import SQLAlchemyUserDatabase
from fps.exceptions import RedirectException  # type: ignore
from fps.logging import get_configured_logger  # type: ignore
from httpx_oauth.clients.github import GitHubOAuth2  # type: ignore
from starlette.requests import Request

from .config import get_auth_config
from .db import User, get_user_db, secret

logger = get_configured_logger("auth")


class NoAuthTransport(Transport):
    scheme = None  # type: ignore


class NoAuthStrategy(Strategy, Generic[models.UP, models.ID]):
    async def read_token(
        self, token: Optional[str], user_manager: BaseUserManager[models.UP, models.ID]
    ) -> Optional[models.UP]:
        active_user = await user_manager.user_db.get_by_email(get_auth_config().global_email)
        return active_user


class GitHubTransport(CookieTransport):
    async def get_login_response(self, token: str, response: Response):
        await super().get_login_response(token, response)
        response.status_code = status.HTTP_302_FOUND
        response.headers["Location"] = "/lab"


noauth_transport = NoAuthTransport()
cookie_transport = CookieTransport(cookie_secure=get_auth_config().cookie_secure)
github_transport = GitHubTransport()


def get_noauth_strategy() -> NoAuthStrategy:
    return NoAuthStrategy()


def get_jwt_strategy() -> JWTStrategy:
    return JWTStrategy(secret=secret, lifetime_seconds=None)  # type: ignore


noauth_authentication = AuthenticationBackend(
    name="noauth",
    transport=noauth_transport,
    get_strategy=get_noauth_strategy,
)
cookie_authentication = AuthenticationBackend(
    name="cookie",
    transport=cookie_transport,
    get_strategy=get_jwt_strategy,
)
github_cookie_authentication = AuthenticationBackend(
    name="github",
    transport=github_transport,
    get_strategy=get_jwt_strategy,
)
github_authentication = GitHubOAuth2(
    get_auth_config().client_id, get_auth_config().client_secret.get_secret_value()
)


class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
    async def on_after_register(self, user: User, request: Optional[Request] = None):
        for oauth_account in user.oauth_accounts:
            if oauth_account.oauth_name == "github":
                async with httpx.AsyncClient() as client:
                    r = (
                        await client.get(f"https://api.github.com/user/{oauth_account.account_id}")
                    ).json()

                await self.user_db.update(
                    user,
                    dict(
                        anonymous=False,
                        username=r["login"],
                        color=None,
                        avatar=r["avatar_url"],
                        is_active=True,
                    ),
                )


def get_user_manager(user_db: SQLAlchemyUserDatabase = Depends(get_user_db)):
    yield UserManager(user_db)


async def get_enabled_backends(auth_config=Depends(get_auth_config)):
    if auth_config.mode == "noauth" and not auth_config.collaborative:
        return [noauth_authentication, github_cookie_authentication]
    else:
        return [cookie_authentication, github_cookie_authentication]


fapi_users = FastAPIUsers[User, uuid.UUID](
    get_user_manager,
    [noauth_authentication, cookie_authentication, github_cookie_authentication],
)


async def create_guest(user_db, auth_config):
    # workspace and settings are copied from global user
    # but this is a new user
    global_user = await user_db.get_by_email(auth_config.global_email)
    user_id = str(uuid.uuid4())
    guest = dict(
        id=user_id,
        anonymous=True,
        email=f"{user_id}@jupyter.com",
        username=f"{user_id}@jupyter.com",
        hashed_password="",
        workspace=global_user.workspace,
        settings=global_user.settings,
    )
    return await user_db.create(guest)


async def current_user(
    response: Response,
    token: Optional[str] = None,
    user: User = Depends(
        fapi_users.current_user(optional=True, get_enabled_backends=get_enabled_backends)
    ),
    user_db=Depends(get_user_db),
    auth_config=Depends(get_auth_config),
):
    active_user = user

    if auth_config.collaborative:
        if not active_user and auth_config.mode == "noauth":
            active_user = await create_guest(user_db, auth_config)
            await cookie_authentication.login(get_jwt_strategy(), active_user, response)

        elif not active_user and auth_config.mode == "token":
            global_user = await user_db.get_by_email(auth_config.global_email)
            if global_user and global_user.hashed_password == token:
                active_user = await create_guest(user_db, auth_config)
                await cookie_authentication.login(get_jwt_strategy(), active_user, response)
    else:
        if auth_config.mode == "token":
            global_user = await user_db.get_by_email(auth_config.global_email)
            if global_user and global_user.hashed_password == token:
                active_user = global_user
                await cookie_authentication.login(get_jwt_strategy(), active_user, response)

    if active_user:
        return active_user

    elif auth_config.login_url:
        raise RedirectException(auth_config.login_url)

    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)



from uuid import uuid4

from fastapi import APIRouter, Depends
from fps.config import get_config  # type: ignore
from fps.hooks import register_router  # type: ignore
from fps.logging import get_configured_logger  # type: ignore
from fps_uvicorn.config import UvicornConfig  # type: ignore
from sqlalchemy import select  # type: ignore

from .backends import (
    cookie_authentication,
    current_user,
    fapi_users,
    github_authentication,
    github_cookie_authentication,
)
from .config import get_auth_config
from .db import Session, User, UserDb, create_db_and_tables, secret
from .models import UserCreate, UserRead, UserUpdate

logger = get_configured_logger("auth")

uvicorn_config = get_config(UvicornConfig)

router = APIRouter()


@router.on_event("startup")
async def startup():
    await create_db_and_tables()

    auth_config = get_auth_config()

    async with UserDb() as user_db:

        global_user = await user_db.get_by_email(auth_config.global_email)

        if global_user:
            await user_db.update(global_user, {"hashed_password": auth_config.token})

        else:
            global_user = dict(
                id=uuid4(),
                anonymous=True,
                email=auth_config.global_email,
                username=auth_config.global_email,
                hashed_password=auth_config.token,
                is_superuser=False,
                is_active=False,
                is_verified=True,
                workspace="{}",
                settings="{}",
            )
            await user_db.create(global_user)

    if auth_config.mode == "token":
        logger.info("")
        logger.info("To access the server, copy and paste this URL:")
        logger.info(
            f"http://{uvicorn_config.host}:{uvicorn_config.port}/?token={auth_config.token}"
        )
        logger.info("")


@router.get("/auth/users")
async def get_users(user: UserRead = Depends(current_user)):
    async with Session() as session:
        statement = select(User)
        users = (await session.execute(statement)).unique().all()
    return [usr.User for usr in users if usr.User.is_active]


# redefine GET /me because we want our current_user dependency
# it is first defined in users_router and so it wins over the one in fapi_users.get_users_router
users_router = APIRouter()


@users_router.get("/me")
async def get_me(user: UserRead = Depends(current_user)):
    return user


users_router.include_router(fapi_users.get_users_router(UserRead, UserUpdate))

# Cookie based auth login and logout
r_cookie_auth = register_router(fapi_users.get_auth_router(cookie_authentication), prefix="/auth")
r_register = register_router(fapi_users.get_register_router(UserRead, UserCreate), prefix="/auth")
r_user = register_router(users_router, prefix="/auth/user")

# GitHub OAuth register router
r_github = register_router(
    fapi_users.get_oauth_router(github_authentication, github_cookie_authentication, secret),
    prefix="/auth/github",
)

r = register_router(router)
